@@ -1,5 +1,5 @@
 /*
- * $RCSId: xc/lib/fontconfig/fc-cache/fc-cache.c,v 1.8tsi Exp $
+ * fontconfig/fc-cache/fc-cache.c
  *
  * Copyright © 2002 Keith Packard
  *
@@ -7,23 +7,21 @@
  * documentation for any purpose is hereby granted without fee, provided that
  * the above copyright notice appear in all copies and that both that
  * copyright notice and this permission notice appear in supporting
- * documentation, and that the name of Keith Packard not be used in
+ * documentation, and that the name of the author(s) not be used in
  * advertising or publicity pertaining to distribution of the software without
- * specific, written prior permission.  Keith Packard makes no
+ * specific, written prior permission.  The authors make no
  * representations about the suitability of this software for any purpose.  It
  * is provided "as is" without express or implied warranty.
  *
- * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
  * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+ * EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "../fc-arch/fcarch.h"
-
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #else
@@ -69,10 +67,11 @@
 const struct option longopts[] = {
     {"force", 0, 0, 'f'},
     {"really-force", 0, 0, 'r'},
+    {"sysroot", required_argument, 0, 'y'},
     {"system-only", 0, 0, 's'},
     {"version", 0, 0, 'V'},
     {"verbose", 0, 0, 'v'},
-    {"help", 0, 0, '?'},
+    {"help", 0, 0, 'h'},
     {NULL,0,0,0},
 };
 #else
@@ -83,40 +82,43 @@ extern int optind, opterr, optopt;
 #endif
 
 static void
-usage (char *program)
+usage (char *program, int error)
 {
+    FILE *file = error ? stderr : stdout;
 #if HAVE_GETOPT_LONG
-    fprintf (stderr, "usage: %s [-frsvV?] [--force|--really-force] [--system-only] [--verbose] [--version] [--help] [dirs]\n",
+    fprintf (file, "usage: %s [-frsvVh] [-y SYSROOT] [--force|--really-force] [--sysroot=SYSROOT] [--system-only] [--verbose] [--version] [--help] [dirs]\n",
 	     program);
 #else
-    fprintf (stderr, "usage: %s [-frsvV?] [dirs]\n",
+    fprintf (file, "usage: %s [-frsvVh] [-y SYSROOT] [dirs]\n",
 	     program);
 #endif
-    fprintf (stderr, "Build font information caches in [dirs]\n"
+    fprintf (file, "Build font information caches in [dirs]\n"
 	     "(all directories in font configuration by default).\n");
-    fprintf (stderr, "\n");
+    fprintf (file, "\n");
 #if HAVE_GETOPT_LONG
-    fprintf (stderr, "  -f, --force          scan directories with apparently valid caches\n");
-    fprintf (stderr, "  -r, --really-force   erase all existing caches, then rescan\n");
-    fprintf (stderr, "  -s, --system-only    scan system-wide directories only\n");
-    fprintf (stderr, "  -v, --verbose        display status information while busy\n");
-    fprintf (stderr, "  -V, --version        display font config version and exit\n");
-    fprintf (stderr, "  -?, --help           display this help and exit\n");
+    fprintf (file, "  -f, --force              scan directories with apparently valid caches\n");
+    fprintf (file, "  -r, --really-force       erase all existing caches, then rescan\n");
+    fprintf (file, "  -s, --system-only        scan system-wide directories only\n");
+    fprintf (file, "  -y, --sysroot=SYSROOT    prepend SYSROOT to all paths for scanning\n");
+    fprintf (file, "  -v, --verbose            display status information while busy\n");
+    fprintf (file, "  -V, --version            display font config version and exit\n");
+    fprintf (file, "  -h, --help               display this help and exit\n");
 #else
-    fprintf (stderr, "  -f         (force)   scan directories with apparently valid caches\n");
-    fprintf (stderr, "  -r,   (really force) erase all existing caches, then rescan\n");
-    fprintf (stderr, "  -s         (system)  scan system-wide directories only\n");
-    fprintf (stderr, "  -v         (verbose) display status information while busy\n");
-    fprintf (stderr, "  -V         (version) display font config version and exit\n");
-    fprintf (stderr, "  -?         (help)    display this help and exit\n");
+    fprintf (file, "  -f         (force)   scan directories with apparently valid caches\n");
+    fprintf (file, "  -r,   (really force) erase all existing caches, then rescan\n");
+    fprintf (file, "  -s         (system)  scan system-wide directories only\n");
+    fprintf (file, "  -y SYSROOT (sysroot) prepend SYSROOT to all paths for scanning\n");
+    fprintf (file, "  -v         (verbose) display status information while busy\n");
+    fprintf (file, "  -V         (version) display font config version and exit\n");
+    fprintf (file, "  -h         (help)    display this help and exit\n");
 #endif
-    exit (1);
+    exit (error);
 }
 
 static FcStrSet *processed_dirs;
 
 static int
-scanDirs (FcStrList *list, FcConfig *config, FcBool force, FcBool really_force, FcBool verbose)
+scanDirs (FcStrList *list, FcConfig *config, FcBool force, FcBool really_force, FcBool verbose, FcBool recursive, int *changed, FcStrSet *updateDirs)
 {
     int		    ret = 0;
     const FcChar8   *dir;
@@ -135,53 +137,37 @@ scanDirs (FcStrList *list, FcConfig *config, FcBool force, FcBool really_force,
     {
 	if (verbose)
 	{
-	    printf ("%s: ", dir);
+	    if (!recursive)
+		printf ("Re-scanning %s: ", dir);
+	    else
+		printf ("%s: ", dir);
 	    fflush (stdout);
 	}
 	
-	if (!dir)
-	{
-	    if (verbose)
-		printf ("skipping, no such directory\n");
-	    continue;
-	}
-	
-	if (FcStrSetMember (processed_dirs, dir))
+	if (recursive && FcStrSetMember (processed_dirs, dir))
 	{
 	    if (verbose)
 		printf ("skipping, looped directory detected\n");
 	    continue;
 	}
 
-	if (access ((char *) dir, W_OK) < 0)
+	if (stat ((char *) dir, &statb) == -1)
 	{
 	    switch (errno) {
 	    case ENOENT:
 	    case ENOTDIR:
 		if (verbose)
 		    printf ("skipping, no such directory\n");
-		continue;
-	    case EACCES:
-	    case EROFS:
-		/* That's ok, caches go to /var anyway. */
-		/* Ideally we'd do an access on the hashed_name. */
-		/* But we hid that behind an abstraction barrier. */
 		break;
 	    default:
 		fprintf (stderr, "\"%s\": ", dir);
 		perror ("");
 		ret++;
-
-		continue;
+		break;
 	    }
-	}
-	if (stat ((char *) dir, &statb) == -1)
-	{
-	    fprintf (stderr, "\"%s\": ", dir);
-	    perror ("");
-	    ret++;
 	    continue;
 	}
+
 	if (!S_ISDIR (statb.st_mode))
 	{
 	    fprintf (stderr, "\"%s\": not a directory, skipping\n", dir);
@@ -201,7 +187,13 @@ scanDirs (FcStrList *list, FcConfig *config, FcBool force, FcBool really_force,
 	
 	if (!cache)
 	{
-	    cache = FcDirCacheRead (dir, FcTrue, config);
+	    if (!recursive)
+		cache = FcDirCacheRescan (dir, config);
+	    else
+	    {
+		(*changed)++;
+		cache = FcDirCacheRead (dir, FcTrue, config);
+	    }
 	    if (!cache)
 	    {
 		fprintf (stderr, "%s: error scanning\n", dir);
@@ -213,13 +205,13 @@ scanDirs (FcStrList *list, FcConfig *config, FcBool force, FcBool really_force,
 	if (was_valid)
 	{
 	    if (verbose)
-		printf ("skipping, %d fonts, %d dirs\n",
+		printf ("skipping, existing cache is valid: %d fonts, %d dirs\n",
 			FcCacheNumFont (cache), FcCacheNumSubdir (cache));
 	}
 	else
 	{
 	    if (verbose)
-		printf ("caching, %d fonts, %d dirs\n", 
+		printf ("caching, new cache contents: %d fonts, %d dirs\n", 
 			FcCacheNumFont (cache), FcCacheNumSubdir (cache));
 
 	    if (!FcDirCacheValid (dir))
@@ -229,127 +221,39 @@ scanDirs (FcStrList *list, FcConfig *config, FcBool force, FcBool really_force,
 		ret++;
 	    }
 	}
-	
-	subdirs = FcStrSetCreate ();
-	if (!subdirs)
-	{
-	    fprintf (stderr, "%s: Can't create subdir set\n", dir);
-	    ret++;
-	    FcDirCacheUnload (cache);
-	    continue;
-	}
-	for (i = 0; i < FcCacheNumSubdir (cache); i++)
-	    FcStrSetAdd (subdirs, FcCacheSubdir (cache, i));
-	
-	FcDirCacheUnload (cache);
-	
-	sublist = FcStrListCreate (subdirs);
-	FcStrSetDestroy (subdirs);
-	if (!sublist)
-	{
-	    fprintf (stderr, "%s: Can't create subdir list\n", dir);
-	    ret++;
-	    continue;
-	}
-	FcStrSetAdd (processed_dirs, dir);
-	ret += scanDirs (sublist, config, force, really_force, verbose);
-    }
-    FcStrListDone (list);
-    return ret;
-}
 
-static FcBool
-cleanCacheDirectory (FcConfig *config, FcChar8 *dir, FcBool verbose)
-{
-    DIR		*d;
-    struct dirent *ent;
-    FcChar8	*dir_base;
-    FcBool	ret = FcTrue;
-    FcBool	remove;
-    FcCache	*cache;
-    struct stat	file_stat;
-    struct stat	target_stat;
-
-    dir_base = FcStrPlus (dir, (FcChar8 *) "/");
-    if (!dir_base)
-    {
-	fprintf (stderr, "%s: out of memory\n", dir);
-	return FcFalse;
-    }
-    if (access ((char *) dir, W_OK|X_OK) != 0)
-    {
-	if (verbose)
-	    printf ("%s: not cleaning unwritable cache directory\n", dir);
-	FcStrFree (dir_base);
-	return FcTrue;
-    }
-    if (verbose)
-	printf ("%s: cleaning cache directory\n", dir);
-    d = opendir ((char *) dir);
-    if (!d)
-    {
-	perror ((char *) dir);
-	FcStrFree (dir_base);
-	return FcFalse;
-    }
-    while ((ent = readdir (d)))
-    {
-	FcChar8	*file_name;
-	const FcChar8	*target_dir;
-
-	if (ent->d_name[0] == '.')
-	    continue;
-	/* skip cache files for different architectures and */
-	/* files which are not cache files at all */
-	if (strlen(ent->d_name) != 32 + strlen ("-" FC_ARCHITECTURE FC_CACHE_SUFFIX) ||
-	    strcmp(ent->d_name + 32, "-" FC_ARCHITECTURE FC_CACHE_SUFFIX))
-	    continue;
-	
-	file_name = FcStrPlus (dir_base, (FcChar8 *) ent->d_name);
-	if (!file_name)
+	if (recursive)
 	{
-	    fprintf (stderr, "%s: allocation failure\n", dir);
-	    ret = FcFalse;
-	    break;
-	}
-	cache = FcDirCacheLoadFile (file_name, &file_stat);
-	if (!cache)
-	{
-	    fprintf (stderr, "%s: invalid cache file: %s\n", dir, ent->d_name);
-	    FcStrFree (file_name);
-	    ret = FcFalse;
-	    continue;
-	}
-	target_dir = FcCacheDir (cache);
-	remove = FcFalse;
-	if (stat ((char *) target_dir, &target_stat) < 0)
-	{
-	    if (verbose)
-		printf ("%s: %s: missing directory: %s \n",
-			dir, ent->d_name, target_dir);
-	    remove = FcTrue;
-	}
-	else if (target_stat.st_mtime > file_stat.st_mtime)
-	{
-	    if (verbose)
-		printf ("%s: %s: cache outdated: %s\n",
-			dir, ent->d_name, target_dir);
-	    remove = FcTrue;
-	}
-	if (remove)
-	{
-	    if (unlink ((char *) file_name) < 0)
+	    subdirs = FcStrSetCreate ();
+	    if (!subdirs)
 	    {
-		perror ((char *) file_name);
-		ret = FcFalse;
+		fprintf (stderr, "%s: Can't create subdir set\n", dir);
+		ret++;
+		FcDirCacheUnload (cache);
+		continue;
 	    }
+	    for (i = 0; i < FcCacheNumSubdir (cache); i++)
+		FcStrSetAdd (subdirs, FcCacheSubdir (cache, i));
+	    if (updateDirs && FcCacheNumSubdir (cache) > 0)
+		FcStrSetAdd (updateDirs, dir);
+	
+	    FcDirCacheUnload (cache);
+	
+	    sublist = FcStrListCreate (subdirs);
+	    FcStrSetDestroy (subdirs);
+	    if (!sublist)
+	    {
+		fprintf (stderr, "%s: Can't create subdir list\n", dir);
+		ret++;
+		continue;
+	    }
+	    FcStrSetAdd (processed_dirs, dir);
+	    ret += scanDirs (sublist, config, force, really_force, verbose, recursive, changed, updateDirs);
+	    FcStrListDone (sublist);
 	}
-	FcDirCacheUnload (cache);
-        FcStrFree (file_name);
+	else
+	    FcDirCacheUnload (cache);
     }
-    
-    closedir (d);
-    FcStrFree (dir_base);
     return ret;
 }
 
@@ -364,7 +268,7 @@ cleanCacheDirectories (FcConfig *config, FcBool verbose)
 	return FcFalse;
     while ((cache_dir = FcStrListNext (cache_dirs)))
     {
-	if (!cleanCacheDirectory (config, cache_dir, verbose))
+	if (!FcDirCacheClean (cache_dir, verbose))
 	{
 	    ret = FcFalse;
 	    break;
@@ -377,22 +281,24 @@ cleanCacheDirectories (FcConfig *config, FcBool verbose)
 int
 main (int argc, char **argv)
 {
-    FcStrSet	*dirs;
+    FcStrSet	*dirs, *updateDirs;
     FcStrList	*list;
     FcBool    	verbose = FcFalse;
     FcBool	force = FcFalse;
     FcBool	really_force = FcFalse;
     FcBool	systemOnly = FcFalse;
     FcConfig	*config;
+    FcChar8     *sysroot = NULL;
     int		i;
+    int		changed;
     int		ret;
 #if HAVE_GETOPT_LONG || HAVE_GETOPT
     int		c;
 
 #if HAVE_GETOPT_LONG
-    while ((c = getopt_long (argc, argv, "frsVv?", longopts, NULL)) != -1)
+    while ((c = getopt_long (argc, argv, "frsy:Vvh", longopts, NULL)) != -1)
 #else
-    while ((c = getopt (argc, argv, "frsVv?")) != -1)
+    while ((c = getopt (argc, argv, "frsy:Vvh")) != -1)
 #endif
     {
 	switch (c) {
@@ -405,6 +311,9 @@ main (int argc, char **argv)
 	case 's':
 	    systemOnly = FcTrue;
 	    break;
+	case 'y':
+	    sysroot = FcStrCopy ((const FcChar8 *)optarg);
+	    break;
 	case 'V':
 	    fprintf (stderr, "fontconfig version %d.%d.%d\n", 
 		     FC_MAJOR, FC_MINOR, FC_REVISION);
@@ -412,8 +321,10 @@ main (int argc, char **argv)
 	case 'v':
 	    verbose = FcTrue;
 	    break;
+	case 'h':
+	    usage (argv[0], 0);
 	default:
-	    usage (argv[0]);
+	    usage (argv[0], 1);
 	}
     }
     i = optind;
@@ -423,7 +334,16 @@ main (int argc, char **argv)
 
     if (systemOnly)
 	FcConfigEnableHome (FcFalse);
-    config = FcInitLoadConfig ();
+    if (sysroot)
+    {
+	FcConfigSetSysRoot (NULL, sysroot);
+	FcStrFree (sysroot);
+	config = FcConfigGetCurrent();
+    }
+    else
+    {
+	config = FcInitLoadConfig ();
+    }
     if (!config)
     {
 	fprintf (stderr, "%s: Can't init font config library\n", argv[0]);
@@ -459,13 +379,33 @@ main (int argc, char **argv)
 	fprintf(stderr, "Cannot malloc\n");
 	return 1;
     }
-	
-    ret = scanDirs (list, config, force, really_force, verbose);
+
+    updateDirs = FcStrSetCreate ();
+    changed = 0;
+    ret = scanDirs (list, config, force, really_force, verbose, FcTrue, &changed, updateDirs);
+    /* Update the directory cache again to avoid the race condition as much as possible */
+    FcStrListDone (list);
+    list = FcStrListCreate (updateDirs);
+    if (list)
+    {
+	ret += scanDirs (list, config, FcTrue, really_force, verbose, FcFalse, &changed, NULL);
+	FcStrListDone (list);
+    }
+    FcStrSetDestroy (updateDirs);
+
+    /*
+     * Try to create CACHEDIR.TAG anyway.
+     * This expects the fontconfig cache directory already exists.
+     * If it doesn't, it won't be simply created.
+     */
+    FcCacheCreateTagFile (config);
 
     FcStrSetDestroy (processed_dirs);
 
     cleanCacheDirectories (config, verbose);
 
+    FcConfigDestroy (config);
+    FcFini ();
     /* 
      * Now we need to sleep a second  (or two, to be extra sure), to make
      * sure that timestamps for changes after this run of fc-cache are later
@@ -473,8 +413,9 @@ main (int argc, char **argv)
      * sleep(3) can't be interrupted by a signal here -- this isn't in the
      * library, and there aren't any signals flying around here.
      */
-    FcConfigDestroy (config);
-    sleep (2);
+    /* the resolution of mtime on FAT is 2 seconds */
+    if (changed)
+	sleep (2);
     if (verbose)
 	printf ("%s: %s\n", argv[0], ret ? "failed" : "succeeded");
     return ret;
